
NAME_TABLE = 'NAME_TABLE'
PARAMETER_TABLE = 'PARAMETER_TABLE'
PORT_TABLE = 'PORT_TABLE'
STATIC_VAR_TABLE = 'STATIC_VAR_TABLE'

ALLOWED_TYPES = 'ALLOWED_TYPES'
ARRAY = 'ARRAY'
ARRAY_BOUNDS = 'ARRAY_BOUNDS'
C_FUNCTION_NAME = 'C_FUNCTION_NAME'
DATA_TYPE = 'DATA_TYPE'
DEFAULT_TYPE = 'DEFAULT_TYPE'
DEFAULT_VALUE = 'DEFAULT_VALUE'
DESCRIPTION = 'DESCRIPTION'
DIRECTION = 'DIRECTION'
LIMITS = 'LIMITS'
NULL_ALLOWED = 'NULL_ALLOWED'
PARAMETER_NAME = 'PARAMETER_NAME'
PORT_NAME = 'PORT_NAME'
SPICE_MODEL_NAME = 'SPICE_MODEL_NAME'
STATIC_VAR_NAME = 'STATIC_VAR_NAME'
